// g++ tcl_rbtree.C -Wall -shared -o rbtree.so -O

// All code below was written by Philip Smolen, copyright 2000, and
// is released free under the GNU Public License. 
// Contact info:  psmole19@idt.net.


#include <tcl.h>
#include <string.h>

#include "rbtree.h"

//--------------------------------------------------------------------------
// internal data structures
//--------------------------------------------------------------------------

class tclObject
{
private:
  Tcl_Obj *_obj;
  void incrRefCount() { if (_obj) Tcl_IncrRefCount(_obj); }
  void decrRefCount() { if (_obj) Tcl_DecrRefCount(_obj); }

public:
  operator Tcl_Obj * () const { return _obj; }

  tclObject() : _obj(NULL) {}
  tclObject(const tclObject &other) : _obj(other._obj) { incrRefCount(); }
  tclObject(Tcl_Obj *obj) : _obj(obj) { incrRefCount(); }
  ~tclObject() { decrRefCount(); }
  tclObject &operator = (const tclObject &other) 
    { decrRefCount(); _obj = other._obj; incrRefCount(); return *this; } 
};

class tclRBTree;

class tclRBNode
{
private: 
  const tclRBTree *_owner;
  tclObject _key, _data;
public:
  tclRBNode(tclRBTree *owner, const tclObject &key, const tclObject &data) :
    _owner(owner), _key(key), _data(data) {}
  tclRBNode() : _owner(NULL) {}
  tclRBNode(Tcl_Obj *obj) : _owner(NULL), _key(obj) {}
  
  tclObject const &getKey() const { return _key; }
  tclObject const &getData() const { return _data; }

  bool operator < (const tclRBNode &other) const;
  bool operator == (const tclRBNode &other) const;

  void makeDataUnique() const
    {
      if (Tcl_IsShared((Tcl_Obj *)_data))
	{
	  ((tclRBNode *)this)->_data = Tcl_DuplicateObj(_data);
	}
    }
};

typedef rbtreeType<tclRBNode> objTree;
typedef objTree::iteratorType objTreeIterator;

class tclRBTree
{
private:
  Tcl_Obj *getKey(objTreeIterator i) 
    { return i?(Tcl_Obj *)(i->getData().getKey()):NULL; }
  const Tcl_Interp *_interp;
  objTree _tree;
public:
  int compare(Tcl_Obj *a, Tcl_Obj *b) const;
  Tcl_Obj *contentsAsList() const;

  objTree &getTree() { return _tree; }

  bool existsCmd(Tcl_Obj *key) { return _tree.exists(key); }
  void insertCmd(Tcl_Obj *key, Tcl_Obj *data) 
    { _tree.add(tclRBNode(this, key, data)); }
  bool deleteCmd(Tcl_Obj *key) 
    { return _tree.remove(tclRBNode(this, key, NULL)); }
  Tcl_Obj *closestCmd(bool before, Tcl_Obj *key) 
    {
      return getKey(_tree.find(tclRBNode(this, key, NULL), 
			       before?objTree::before:objTree::after));
    }
  Tcl_Obj *previousCmd(Tcl_Obj *key) 
    {
      return getKey(_tree.prev(_tree.find(tclRBNode(this, key, NULL), 
					  objTree::after)));
    }
  Tcl_Obj *nextCmd(Tcl_Obj *key) 
    {
      return getKey(_tree.next(_tree.find(tclRBNode(this, key, NULL), 
					  objTree::before)));
    }
  Tcl_Obj *keysCmd()
    {
      Tcl_Obj *result = Tcl_NewObj();
      for (objTreeIterator i = _tree.first(); i; i = _tree.next(i))
	{
	  Tcl_ListObjAppendElement(NULL, result, i->getData().getKey());
	}
      return result;
    }
  Tcl_Obj *valueCmd(Tcl_Obj *key)
    {
      objTreeIterator found = _tree.find(tclRBNode(this, key, NULL), 
					 objTree::exact);
      if (found)
	{
	  return found->getData().getData();
	}
      else
	{ 
	  return NULL; 
	}
    } 

  tclRBTree(Tcl_Interp *interp) : _interp(interp) {}
};

inline bool tclRBNode::operator < (const tclRBNode &other) const
{
  return _owner->compare(_key, other._key) < 0;
}

inline bool tclRBNode::operator == (const tclRBNode &other) const
{
  return _owner->compare(_key, other._key) == 0;
}

// This is called a whole lot.  Often with the same data, or with 
// a and b reversed, two or more times in a row.
int tclRBTree::compare(Tcl_Obj *a, Tcl_Obj *b) const
{
  int aLen;
  char *aData = Tcl_GetStringFromObj(a, &aLen);
  int bLen;
  char *bData = Tcl_GetStringFromObj(b, &bLen);
  //  cout<<"compare("<<aData<<','<<bData<<") -> ";
  int len = (aLen<bLen)?aLen:bLen;
  int cmp = memcmp(aData, bData, len);
  if (cmp == 0)
    {
      //      cout<<(aLen-bLen)<<" by the first case.\n";
      return aLen-bLen;
    }
  else
    { 
      //      cout<<cmp<<" by the second case.\n";
      return cmp;
    }
}

Tcl_Obj *tclRBTree::contentsAsList() const
{
  Tcl_Obj *list = Tcl_NewObj();
  for (objTreeIterator i = _tree.first(); i; i = _tree.next(i))
    {
      Tcl_ListObjAppendElement(NULL, list, i->getData().getKey());
      Tcl_ListObjAppendElement(NULL, list, i->getData().getData());
    }
  return list;
}

//--------------------------------------------------------------------------
// tree data type
//--------------------------------------------------------------------------

static void FreeTreeInternalRep(Tcl_Obj *objPtr);
static void DupTreeInternalRep(Tcl_Obj *srcPtr, Tcl_Obj *copyPtr);
static int SetTreeFromAny(Tcl_Interp *interp, Tcl_Obj *objPtr);
static void UpdateStringOfTree(Tcl_Obj *objPtr);

static Tcl_ObjType treeType = {
  "rbtree",				/* name */
  FreeTreeInternalRep,                  /* freeIntRepProc */
  DupTreeInternalRep,		        /* dupIntRepProc */
  UpdateStringOfTree,			/* updateStringProc */
  SetTreeFromAny			/* setFromAnyProc */
};

static Tcl_Obj *newTreeObj(Tcl_Interp *interp) {
  Tcl_Obj *objPtr;
  
  objPtr = Tcl_NewObj();
  objPtr->bytes = NULL;
  objPtr->internalRep.otherValuePtr = new tclRBTree(interp);
  objPtr->typePtr = &treeType;
  return objPtr;
}

static void FreeTreeInternalRep(Tcl_Obj *objPtr)
{
  delete (tclRBTree *)objPtr->internalRep.otherValuePtr;
  objPtr->typePtr = NULL;
}

static void DupTreeInternalRep(Tcl_Obj *srcPtr, Tcl_Obj *copyPtr)
{
  copyPtr->internalRep.otherValuePtr = 
    new tclRBTree(*(tclRBTree *)srcPtr->internalRep.otherValuePtr);
  copyPtr->typePtr = &treeType;
}

static int SetTreeFromAny(Tcl_Interp *interp, Tcl_Obj *objPtr)
{
  Tcl_Obj *asList = Tcl_DuplicateObj(objPtr);
  tclObject gc = asList;
  int count;
  if (Tcl_ListObjLength(interp, objPtr, &count) != TCL_OK)
    {
      return TCL_ERROR;
    }
  if (count % 2)
    {
      Tcl_AppendResult(interp, "Expecting an even number of items", NULL);
      return TCL_ERROR;
    }

  if (objPtr->typePtr && objPtr->typePtr->freeIntRepProc) 
    {
      objPtr->typePtr->freeIntRepProc(objPtr);
    }

  tclRBTree *tree = new tclRBTree(interp);
  for(int i = 0; i < count; i += 2)
    {
      Tcl_Obj *key = NULL;
      Tcl_ListObjIndex(NULL, asList, i, &key);
      Tcl_Obj *data = NULL;
      Tcl_ListObjIndex(NULL, asList, i+1, &data);
      tree->insertCmd(key, data);
    }

  objPtr->internalRep.otherValuePtr = tree;
  objPtr->typePtr = &treeType;
  return TCL_OK;
}

static void UpdateStringOfTree(Tcl_Obj *objPtr)
{
  tclRBTree *tree = (tclRBTree *)objPtr->internalRep.otherValuePtr;
  Tcl_Obj *list = tree->contentsAsList();
  char *str = Tcl_GetStringFromObj(list, &objPtr->length);
  objPtr->bytes = Tcl_Alloc(objPtr->length+1);
  memcpy(objPtr->bytes, str, objPtr->length+1);
  Tcl_IncrRefCount(list);
  Tcl_DecrRefCount(list);
}


//--------------------------------------------------------------------------
// 
//--------------------------------------------------------------------------

static Tcl_Obj *tclOne = NULL;
static Tcl_Obj *&tclTrue = tclOne;
static Tcl_Obj *tclFalse = NULL;

// Note that we do not hold onto a reference count.  This may become a
// problem when we allow callbacks for the compare command.  We don't
// want someone to delete the variable containing the only reference
// count to our data while we're still manipulating the data.
static tclRBTree *getTree(Tcl_Interp *interp, Tcl_Obj *varName, bool forWrite)
{
  Tcl_Obj *treeObj = Tcl_ObjGetVar2(interp, varName, NULL,
				    TCL_LEAVE_ERR_MSG | TCL_PARSE_PART1);
  if (!treeObj) 
    {
      return NULL;
    }
  if (forWrite && Tcl_IsShared(treeObj))
    { 
      tclObject newTree(Tcl_DuplicateObj(treeObj));
      treeObj = Tcl_ObjSetVar2(interp, varName, NULL, newTree,
			       TCL_LEAVE_ERR_MSG | TCL_PARSE_PART1);
      if (!treeObj)
	{
	  return NULL;
	}
    }
  if (Tcl_ConvertToType(interp, treeObj, &treeType) != TCL_OK)
    {
      return NULL;
    }
  if (forWrite)
    {
      Tcl_InvalidateStringRep(treeObj);
    }
  return (tclRBTree *)treeObj->internalRep.otherValuePtr;
}

static tclObject getTreeObj(Tcl_Interp *interp, 
			    Tcl_Obj *varName, 
			    tclRBTree *&commandData,
			    objTree *&tree)
{
  commandData = NULL;
  tree = NULL;
  Tcl_Obj *treeObj = Tcl_ObjGetVar2(interp, varName, NULL,
				    TCL_LEAVE_ERR_MSG | TCL_PARSE_PART1);
  if (!treeObj) 
    {
      return NULL;
    }
  if (Tcl_ConvertToType(interp, treeObj, &treeType) != TCL_OK)
    {
      return NULL;
    }
  commandData = (tclRBTree *)treeObj->internalRep.otherValuePtr;
  tree = &commandData->getTree();
  return treeObj;
}

static Tcl_Obj *emptyAsNull(Tcl_Obj *obj) 
{
  int len;
  Tcl_GetStringFromObj(obj, &len);
  return len?obj:NULL;
}

static int tree_create_cmd(ClientData dummy, 
			Tcl_Interp *interp,
			int objc,
			Tcl_Obj *CONST objv[]) 
{
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "");
    return TCL_ERROR;
  }
  Tcl_SetObjResult(interp, newTreeObj(interp));
  return TCL_OK;
}

static int tree_previous_cmd(ClientData dummy, 
			     Tcl_Interp *interp,
			     int objc,
			     Tcl_Obj *CONST objv[]) 
{
  if ((objc < 3) || (objc > 4))
    {
      Tcl_WrongNumArgs(interp, 1, objv, "var key ?error_value?");
      return TCL_ERROR;
    }
  tclRBTree *tree = getTree(interp, objv[1], false);
  if (!tree) { return TCL_ERROR; }
  Tcl_Obj *result = tree->previousCmd(objv[2]);
  if (result)
    {
      Tcl_SetObjResult(interp, result);
      return TCL_OK;
    }
  else if (objc == 4)
    {
      Tcl_SetObjResult(interp, objv[3]);
      return TCL_OK;
    }
  else 
    {
      Tcl_AppendResult(interp, "No previous element in tree.", NULL);
      return TCL_ERROR;
    }
}

static int tree_next_cmd(ClientData dummy, 
			 Tcl_Interp *interp,
			 int objc,
			 Tcl_Obj *CONST objv[]) 
{
  if ((objc < 3) || (objc > 4))
    {
      Tcl_WrongNumArgs(interp, 1, objv, "var key ?error_value?");
      return TCL_ERROR;
    }
  tclRBTree *tree = getTree(interp, objv[1], false);
  if (!tree) { return TCL_ERROR; }
  Tcl_Obj *result = tree->nextCmd(objv[2]);
  if (result)
    {
      Tcl_SetObjResult(interp, result);
      return TCL_OK;
    }
  else if (objc == 4)
    {
      Tcl_SetObjResult(interp, objv[3]);
      return TCL_OK;
    }
  else 
    {
      Tcl_AppendResult(interp, "No next element in tree.", NULL);
      return TCL_ERROR;
    }
}

static int tree_closest_cmd(ClientData dummy, 
			 Tcl_Interp *interp,
			 int objc,
			 Tcl_Obj *CONST objv[]) 
{
  if ((objc < 4) || (objc > 5))
    {
      Tcl_WrongNumArgs(interp, 1, objv, "var position key ?error_value?");
      return TCL_ERROR;
    }
  static char *positions[] = {"before", "after", NULL};
  int positionIndex;
  if (Tcl_GetIndexFromObj(interp, objv[2], positions, "position", 0,
			  &positionIndex) != TCL_OK) 
    {
      return TCL_ERROR;
    }
  tclRBTree *tree = getTree(interp, objv[1], false);
  if (!tree) { return TCL_ERROR; }
  Tcl_Obj *result = tree->closestCmd(positionIndex==0, objv[3]);
  if (result)
    {
      Tcl_SetObjResult(interp, result);
      return TCL_OK;
    }
  else if (objc == 5)
    {
      Tcl_SetObjResult(interp, objv[4]);
      return TCL_OK;
    }
  else 
    {
      Tcl_AppendResult(interp, "No such element in tree.", NULL);
      return TCL_ERROR;
    }
}

static int tree_set_cmd(ClientData dummy, 
			Tcl_Interp *interp,
			int objc,
			Tcl_Obj *CONST objv[]) 
{
  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "option var ?arg ...?");
    return TCL_ERROR;
  }

  static char *options[] = {
    "exists",	"insert",	"delete",   "foreach",
    "keys",    	NULL
  };
  enum options {
    SET_EXISTS, SET_INSERT, SET_DELETE, SET_FOREACH, 
    SET_KEYS
  };	  
  int index;
  if (Tcl_GetIndexFromObj(interp, objv[1], options, "option", 0,
			  &index) != TCL_OK) {
    return TCL_ERROR;
  }

  tclRBTree *tree;
  switch (index)
    {
    case SET_EXISTS:
      {
	if (objc != 4)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var key");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], false);
	if (!tree) { return TCL_ERROR; }
	Tcl_SetObjResult(interp, tree->existsCmd(objv[3])?tclTrue:tclFalse);
	return TCL_OK;
      }
    case SET_INSERT:
      {
	if (objc != 4)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var key");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], true);
	if (!tree) { return TCL_ERROR; }
	tree->insertCmd(objv[3], objv[3]);
	return TCL_OK;
      }
    case SET_DELETE:
      {
	if (objc != 4)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var key");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], true);
	if (!tree) { return TCL_ERROR; }
	Tcl_SetObjResult(interp, tree->deleteCmd(objv[3])?tclTrue:tclFalse);
	return TCL_OK;
      }
    case SET_FOREACH:
      {
	if ((objc != 5) && (objc != 7))
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, 
			     "var key_var ?from to? body");
	    return TCL_ERROR;
	  }

	Tcl_Obj *keyVarObj = emptyAsNull(objv[3]);

	Tcl_Obj *fromObj, *toObj, *bodyObj;
	if (objc == 5)
	  {
	    fromObj = NULL;
	    toObj = NULL;
	    bodyObj = objv[4];
	  }
	else
	  { 
	    fromObj = objv[4];
	    toObj = objv[5];
	    bodyObj = objv[6];
	  }

	objTree *realTree;
	tclObject lockedObject = getTreeObj(interp, objv[2], tree, realTree);
	if (!lockedObject) { return TCL_ERROR; }

	// first points to the first item in the tree that we plan to 
	// execute.  pastEnd points to the first item after the last item
	// we plan to execute.  If these are the same, we don't go through
	// the loop at all.  We never execute pastEnd, but we often execute
	// first.
	objTreeIterator first;
	objTreeIterator pastEnd;

	if (fromObj && toObj && (tree->compare(fromObj, toObj) > 0))
	  { // If the starting position is greater than the ending position
	    // then we don't do the loop at all.
	    first = NULL;
	    pastEnd = NULL;
	  }
	else
	  { // Otherwise try to respect the user's request.
	    if (fromObj)
	      { // Find the "closest after" item for this.  That is to say
		// we start on the item if it exists, otherwise we start on
		// the first item past where that item would be.
		first = realTree->find(fromObj, objTree::after);
	      }
	    else
	      { // Start at the beginning.
		first = realTree->first();
	      }
	    if (toObj)
	      { // Find the "next" item for this.  That is to say we stop
		// on the first item after this item, if it exists, or after
		// where this item would be, if it does not exist.
		pastEnd = realTree->find(toObj, objTree::before);
		if (pastEnd) { pastEnd = realTree->next(pastEnd); }
	      }
	    else
	      { // Stop at the end.
		pastEnd = NULL;
	      }
	  }

	for (objTreeIterator i = first; i != pastEnd; i = realTree->next(i))
	  {
	    if (!i)
	      { // This could only be the result of a bad compare function.
		// We have hit the end of the list before hitting the expected
		// last element.
		break;
	      }
	    if (keyVarObj)
	      {
		if (!Tcl_ObjSetVar2(interp, keyVarObj, NULL, 
				    i->getData().getKey(),
				    TCL_LEAVE_ERR_MSG | TCL_PARSE_PART1))
		  {
		    return TCL_ERROR;
		  }
	      }
	    int result = Tcl_EvalObj(interp, bodyObj);
	    if (result == TCL_BREAK) { break; }
	    if ((result != TCL_OK) && (result != TCL_CONTINUE))
	      {
		return result;
	      }
	  }

	Tcl_ResetResult(interp);
	return TCL_OK;
      }
    case SET_KEYS:
      {
	if (objc != 3)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], true);
	if (!tree) { return TCL_ERROR; }
	Tcl_SetObjResult(interp, tree->keysCmd());
	return TCL_OK;
      }
    }
  return TCL_OK;
}
  
static int tree_multiset_cmd(ClientData dummy, 
			     Tcl_Interp *interp,
			     int objc,
			     Tcl_Obj *CONST objv[])
{
  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "option var ?arg ...?");
    return TCL_ERROR;
  }

  static char *options[] = {
    "exists",	"insert",	"delete",   "foreach",
    "keys",    	NULL
  };
  enum options {
    MSET_EXISTS, MSET_INSERT, MSET_DELETE, MSET_FOREACH, 
    MSET_KEYS
  };	  
  int index;
  if (Tcl_GetIndexFromObj(interp, objv[1], options, "option", 0,
			  &index) != TCL_OK) {
    return TCL_ERROR;
  }

  tclRBTree *tree;
  switch (index)
    {
    case MSET_EXISTS:
      {
	if (objc != 4)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var key");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], false);
	if (!tree) { return TCL_ERROR; }
	Tcl_Obj *found = tree->valueCmd(objv[3]);
	if (!found)
	  { // No exact match was found.
	    Tcl_SetObjResult(interp, tclFalse);
	  }
	else
	  { // Return the value.  This should be a count, which will be true
	    // unless the count is 0.
	    Tcl_SetObjResult(interp, found);
	  }
	return TCL_OK;
      }
    case MSET_INSERT:
      {
	if (objc != 4)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var key");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], true);
	if (!tree) { return TCL_ERROR; }
	objTreeIterator i = tree->getTree().find(objv[3], objTree::exact);
	if (i)
	  { // The element exists in the tree.  Increment it's count.
	    i->getData().makeDataUnique();
	    int previousCount;
	    if (Tcl_GetIntFromObj(interp, 
				  i->getData().getData(), 
				  &previousCount) != TCL_OK)
	      {
		return TCL_ERROR;
	      }
	    Tcl_SetIntObj(i->getData().getData(), previousCount+1);
	  }
	else
	  { // The element does not yet exist in the tree.
	    tree->insertCmd(objv[3], tclOne);
	  }
	return TCL_OK;
      }
    case MSET_DELETE:
      {
	if (objc != 4)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var key");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], true);
	if (!tree) { return TCL_ERROR; }
	objTreeIterator i = tree->getTree().find(objv[3], objTree::exact);
	if (i)
	  { // The element exists in the tree.
	    int previousCount;
	    if (Tcl_GetIntFromObj(interp, 
				  i->getData().getData(), 
				  &previousCount) != TCL_OK)
	      {
		return TCL_ERROR;
	      }
	    if (previousCount > 1)
	      { // At least one will be left, so decrement the counter.
		i->getData().makeDataUnique();
		Tcl_SetIntObj(i->getData().getData(), previousCount-1);
		Tcl_SetObjResult(interp, tclTrue);
	      }
	    else
	      { // None are left, so remove the entry from the tree completely.
		// We could do this remove a little more quickly if we had
		// access to the remove method that takes an iterator directly.
		tree->deleteCmd(i->getData().getKey());
		if (previousCount > 0)
		  { // There was one element, which we successfully deleted.
		    Tcl_SetObjResult(interp, tclTrue);
		  }
		else
		  { // It's tempting not to even deal with this case.  This
		    // would only happen if someone other than TREE::MSET
		    // manipulated the tree.  If the element appears to have
		    // a count of 0 or less, the delete fails, but we still
		    // remove the entry.
		    Tcl_SetObjResult(interp, tclFalse);
		  }
	      }
	  }
	else
	  { // The element does not exist in the tree.
	    Tcl_SetObjResult(interp, tclFalse);
	  }
	return TCL_OK;
      }
    case MSET_FOREACH:
      {
	if ((objc != 5) && (objc != 7))
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, 
			     "var key_var ?from to? body");
	    return TCL_ERROR;
	  }

	Tcl_Obj *keyVarObj = emptyAsNull(objv[3]);

	Tcl_Obj *fromObj, *toObj, *bodyObj;
	if (objc == 5)
	  {
	    fromObj = NULL;
	    toObj = NULL;
	    bodyObj = objv[4];
	  }
	else
	  { 
	    fromObj = objv[4];
	    toObj = objv[5];
	    bodyObj = objv[6];
	  }

	objTree *realTree;
	tclObject lockedObject = getTreeObj(interp, objv[2], tree, realTree);
	if (!lockedObject) { return TCL_ERROR; }

	// first points to the first item in the tree that we plan to 
	// execute.  pastEnd points to the first item after the last item
	// we plan to execute.  If these are the same, we don't go through
	// the loop at all.  We never execute pastEnd, but we often execute
	// first.
	objTreeIterator first;
	objTreeIterator pastEnd;

	if (fromObj && toObj && (tree->compare(fromObj, toObj) > 0))
	  { // If the starting position is greater than the ending position
	    // then we don't do the loop at all.
	    first = NULL;
	    pastEnd = NULL;
	  }
	else
	  { // Otherwise try to respect the user's request.
	    if (fromObj)
	      { // Find the "closest after" item for this.  That is to say
		// we start on the item if it exists, otherwise we start on
		// the first item past where that item would be.
		first = realTree->find(fromObj, objTree::after);
	      }
	    else
	      { // Start at the beginning.
		first = realTree->first();
	      }
	    if (toObj)
	      { // Find the "next" item for this.  That is to say we stop
		// on the first item after this item, if it exists, or after
		// where this item would be, if it does not exist.
		pastEnd = realTree->find(toObj, objTree::before);
		if (pastEnd) { pastEnd = realTree->next(pastEnd); }
	      }
	    else
	      { // Stop at the end.
		pastEnd = NULL;
	      }
	  }

	for (objTreeIterator i = first; i != pastEnd; i = realTree->next(i))
	  {
	    if (!i)
	      { // This could only be the result of a bad compare function.
		// We have hit the end of the list before hitting the expected
		// last element.
		break;
	      }

	    // Need to find the number of identical keys.
	    int duplicates;
	    if (Tcl_GetIntFromObj(interp,
				  i->getData().getData(), 
				  &duplicates) != TCL_OK)
	      {
		return TCL_ERROR;
	      }
	    
	    for (;duplicates > 0; duplicates--)
	      {
		if (keyVarObj)
		  {
		    if (!Tcl_ObjSetVar2(interp, keyVarObj, NULL, 
					i->getData().getKey(),
					TCL_LEAVE_ERR_MSG | TCL_PARSE_PART1))
		      {
			return TCL_ERROR;
		      }
		  }
		int result = Tcl_EvalObj(interp, bodyObj);
		if (result == TCL_BREAK) { goto break_from_effective_loop; }
		if ((result != TCL_OK) && (result != TCL_CONTINUE))
		  {
		    return result;
		  }
	      }
	  }
	// The two C++ for loops make one appearent foreach loop.  Break
	// jumps out of the foreach, and therefore out of both fors.
      break_from_effective_loop:

	Tcl_ResetResult(interp);
	return TCL_OK;
      }
    case MSET_KEYS:
      {
	if (objc != 3)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], false);
	if (!tree) { return TCL_ERROR; }
	objTree *realTree = &tree->getTree();
	tclObject result = Tcl_NewObj();
	for (objTreeIterator i = realTree->first(); i; i = realTree->next(i))
	  {
	    // Need to find the key...
	    Tcl_Obj *key = i->getData().getKey();

	    // ...and the number of identical keys.
	    int duplicates;
	    if (Tcl_GetIntFromObj(interp,
				  i->getData().getData(), 
				  &duplicates) != TCL_OK)
	      {
		return TCL_ERROR;
	      }
	    
	    for (;duplicates > 0; duplicates--)
	      {
		Tcl_ListObjAppendElement(NULL, result, key);
	      }
	  }
	Tcl_SetObjResult(interp, result);
	return TCL_OK;
      }
    }
  return TCL_OK;
}
  
static int tree_map_cmd(ClientData dummy, 
			Tcl_Interp *interp,
			int objc,
			Tcl_Obj *CONST objv[]) 
{
  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "option var ?arg ...?");
    return TCL_ERROR;
  }

  static char *options[] = {
    "exists",	"insert",	"delete",   "foreach",
    "keys",	"data",		"value",	NULL
  };
  enum options {
    MAP_EXISTS, MAP_INSERT, MAP_DELETE, MAP_FOREACH, 
    MAP_KEYS, MAP_DATA, MAP_VALUE
  };	  
  
  int index;
  if (Tcl_GetIndexFromObj(interp, objv[1], options, "option", 0,
			  &index) != TCL_OK) {
    return TCL_ERROR;
  }

  tclRBTree *tree;
  switch (index)
    {
    case MAP_EXISTS:
      {
	if (objc != 4)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var key");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], false);
	if (!tree) { return TCL_ERROR; }
	Tcl_SetObjResult(interp, tree->existsCmd(objv[3])?tclTrue:tclFalse);
	return TCL_OK;
      }
    case MAP_INSERT:
      {
	if (objc != 5)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var key data");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], true);
	if (!tree) { return TCL_ERROR; }
	tree->insertCmd(objv[3], objv[4]);
	return TCL_OK;
      }
    case MAP_DELETE:
      {
	if (objc != 4)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var key");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], true);
	if (!tree) { return TCL_ERROR; }
	Tcl_SetObjResult(interp, tree->deleteCmd(objv[3])?tclTrue:tclFalse);
	return TCL_OK;
      }
    case MAP_FOREACH:
      {
	if ((objc != 6) && (objc != 8))
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, 
			     "var key_var data_var ?from to? body");
	    return TCL_ERROR;
	  }

	Tcl_Obj *keyVarObj = emptyAsNull(objv[3]);
	Tcl_Obj *dataVarObj = emptyAsNull(objv[4]);

	Tcl_Obj *fromObj, *toObj, *bodyObj;
	if (objc == 6)
	  {
	    fromObj = NULL;
	    toObj = NULL;
	    bodyObj = objv[5];
	  }
	else
	  { 
	    fromObj = objv[5];
	    toObj = objv[6];
	    bodyObj = objv[7];
	  }

	objTree *realTree;
	tclObject lockedObject = getTreeObj(interp, objv[2], tree, realTree);
	if (!lockedObject) { return TCL_ERROR; }

	// first points to the first item in the tree that we plan to 
	// execute.  pastEnd points to the first item after the last item
	// we plan to execute.  If these are the same, we don't go through
	// the loop at all.  We never execute pastEnd, but we often execute
	// first.
	objTreeIterator first;
	objTreeIterator pastEnd;

	if (fromObj && toObj && (tree->compare(fromObj, toObj) > 0))
	  { // If the starting position is greater than the ending position
	    // then we don't do the loop at all.
	    first = NULL;
	    pastEnd = NULL;
	  }
	else
	  { // Otherwise try to respect the user's request.
	    if (fromObj)
	      { // Find the "closest after" item for this.  That is to say
		// we start on the item if it exists, otherwise we start on
		// the first item past where that item would be.
		first = realTree->find(fromObj, objTree::after);
	      }
	    else
	      { // Start at the beginning.
		first = realTree->first();
	      }
	    if (toObj)
	      { // Find the "next" item for this.  That is to say we stop
		// on the first item after this item, if it exists, or after
		// where this item would be, if it does not exist.
		pastEnd = realTree->find(toObj, objTree::before);
		if (pastEnd) { pastEnd = realTree->next(pastEnd); }
	      }
	    else
	      { // Stop at the end.
		pastEnd = NULL;
	      }
	  }

	for (objTreeIterator i = first; i != pastEnd; i = realTree->next(i))
	  {
	    if (!i)
	      { // This could only be the result of a bad compare function.
		// We have hit the end of the list before hitting the expected
		// last element.
		break;
	      }
	    if (keyVarObj)
	      {
		if (!Tcl_ObjSetVar2(interp, keyVarObj, NULL, 
				    i->getData().getKey(),
				    TCL_LEAVE_ERR_MSG | TCL_PARSE_PART1))
		  {
		    return TCL_ERROR;
		  }
	      }
	    if (dataVarObj)
	      {
		if (!Tcl_ObjSetVar2(interp, dataVarObj, NULL, 
				    i->getData().getData(),
				    TCL_LEAVE_ERR_MSG | TCL_PARSE_PART1))
		  {
		    return TCL_ERROR;
		  }
	      }
	    int result = Tcl_EvalObj(interp, bodyObj);
	    if (result == TCL_BREAK) { break; }
	    if ((result != TCL_OK) && (result != TCL_CONTINUE))
	      {
		return result;
	      }
	  }

	Tcl_ResetResult(interp);
	return TCL_OK;
      }
    case MAP_KEYS:
      {
	if (objc != 3)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], true);
	if (!tree) { return TCL_ERROR; }
	Tcl_SetObjResult(interp, tree->keysCmd());
	return TCL_OK;
      }
    case MAP_DATA:
      {
	if (objc != 3)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], true);
	if (!tree) { return TCL_ERROR; }
	Tcl_Obj *result = Tcl_NewObj();
	for (objTreeIterator i = tree->getTree().first(); 
	     i; 
	     i = tree->getTree().next(i))
	  {
	    Tcl_ListObjAppendElement(NULL, result, i->getData().getData());
	  }
	Tcl_SetObjResult(interp, result);
	return TCL_OK;
      }
    case MAP_VALUE:
      {
	if ((objc < 4) || (objc > 5))
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var key ?error_value?");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], false);
	if (!tree) { return TCL_ERROR; }
	Tcl_Obj *result = tree->valueCmd(objv[3]);
	if (result)
	  { // Found!
	    Tcl_SetObjResult(interp, result);
	    return TCL_OK;
	  }
	else
	  { // Not found in tree.
	    if (objc == 5)
	      { // Use default.
		Tcl_SetObjResult(interp, objv[4]);
		return TCL_OK;
	      }
	    else
	      { // Signal an error.
		Tcl_AppendResult(interp, "No such element in tree.", NULL);
		return TCL_ERROR;
	      }
	  }
      }    
    }
  return TCL_OK;
}

static int tree_multimap_cmd(ClientData dummy, 
			     Tcl_Interp *interp,
			     int objc,
			     Tcl_Obj *CONST objv[])
{
  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "option var ?arg ...?");
    return TCL_ERROR;
  }

  static char *options[] = {
    "exists",	"insert",	"delete",	"foreach",
    "keys",	"data",		"value",	NULL
  };
  enum options {
    MMAP_EXISTS, MMAP_INSERT, MMAP_DELETE, MMAP_FOREACH, 
    MMAP_KEYS, MMAP_DATA, MMAP_VALUE
  };
  int index;
  if (Tcl_GetIndexFromObj(interp, objv[1], options, "option", 0,
			  &index) != TCL_OK) {
    return TCL_ERROR;
  }

  tclRBTree *tree;
  switch (index)
    {
    case MMAP_EXISTS:
      {
	if (objc != 4)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var key");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], false);
	if (!tree) { return TCL_ERROR; }
	objTreeIterator found = tree->getTree().find(objv[3], objTree::exact);
	if (!found)
	  { // No exact match was found.
	    Tcl_SetObjResult(interp, tclFalse);
	  }
	else
	  { // Return the length of the list, which is the number of values for
	    // this key.
	    int count;
	    if (Tcl_ListObjLength(interp, found->getData().getData(), &count)
		!= TCL_OK) 
	      {
		return TCL_ERROR;
	      }
	    Tcl_SetObjResult(interp, Tcl_NewIntObj(count));
	  }
	return TCL_OK;
      }
    case MMAP_INSERT:
      {
	if (objc != 5)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var key data");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], true);
	if (!tree) { return TCL_ERROR; }
	objTreeIterator i = tree->getTree().find(objv[3], objTree::exact);
	if (i)
	  { // The element exists in the tree.  lappend the new data.
	    i->getData().makeDataUnique();
	    if (Tcl_ListObjAppendElement(interp, 
					 i->getData().getData(), 
					 objv[4]) != TCL_OK)
	      {
		return TCL_ERROR;
	      }
	  }
	else
	  { // The element does not yet exist in the tree.
	    tree->insertCmd(objv[3], Tcl_NewListObj(1, &objv[4]));
	  }
	return TCL_OK;
      }
    case MMAP_DELETE:
      //needs work!!!
      {
	if (objc != 4)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var key");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], true);
	if (!tree) { return TCL_ERROR; }
	objTreeIterator i = tree->getTree().find(objv[3], objTree::exact);
	if (i)
	  { // The element exists in the tree.
	    int elementCount;
	    if (Tcl_ListObjLength(interp, 
				  i->getData().getData(), 
				  &elementCount) != TCL_OK)
	      {
		return TCL_ERROR;
	      }
	    if (elementCount > 1)
	      { // At least one will be left, so remove one element.  Remove
		// the last element since that will be fastest.  Since
		// we tell the user that he has no say in the order, we can 
		// choose whichever we like.
		i->getData().makeDataUnique();
		Tcl_ListObjReplace(NULL, i->getData().getData(), 
				   elementCount - 1, 
				   1, 0, NULL);
		Tcl_SetObjResult(interp, tclTrue);
	      }
	    else
	      { // None are left, so remove the entry from the tree completely.
		// We could do this remove a little more quickly if we had
		// access to the remove method that takes an iterator directly.
		tree->deleteCmd(i->getData().getKey());
		if (elementCount > 0)
		  { // There was one element, which we successfully deleted.
		    Tcl_SetObjResult(interp, tclTrue);
		  }
		else
		  { // It's tempting not to even deal with this case.  This
		    // would only happen if someone other than TREE::MMAP
		    // manipulated the tree.  If the data has a length of 0, 
		    // the delete fails, but we still remove the entry.
		    Tcl_SetObjResult(interp, tclFalse);
		  }
	      }
	  }
	else
	  { // The element does not exist in the tree.
	    Tcl_SetObjResult(interp, tclFalse);
	  }
	return TCL_OK;
      }
    case MMAP_FOREACH:
      {
	if ((objc != 6) && (objc != 8))
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, 
			     "var key_var ?from to? body");
	    return TCL_ERROR;
	  }

	Tcl_Obj *keyVarObj = emptyAsNull(objv[3]);
	Tcl_Obj *dataVarObj = emptyAsNull(objv[4]);

	Tcl_Obj *fromObj, *toObj, *bodyObj;
	if (objc == 6)
	  {
	    fromObj = NULL;
	    toObj = NULL;
	    bodyObj = objv[5];
	  }
	else
	  { 
	    fromObj = objv[5];
	    toObj = objv[6];
	    bodyObj = objv[7];
	  }

	objTree *realTree;
	tclObject lockedObject = getTreeObj(interp, objv[2], tree, realTree);
	if (!lockedObject) { return TCL_ERROR; }

	// first points to the first item in the tree that we plan to 
	// execute.  pastEnd points to the first item after the last item
	// we plan to execute.  If these are the same, we don't go through
	// the loop at all.  We never execute pastEnd, but we often execute
	// first.
	objTreeIterator first;
	objTreeIterator pastEnd;

	if (fromObj && toObj && (tree->compare(fromObj, toObj) > 0))
	  { // If the starting position is greater than the ending position
	    // then we don't do the loop at all.
	    first = NULL;
	    pastEnd = NULL;
	  }
	else
	  { // Otherwise try to respect the user's request.
	    if (fromObj)
	      { // Find the "closest after" item for this.  That is to say
		// we start on the item if it exists, otherwise we start on
		// the first item past where that item would be.
		first = realTree->find(fromObj, objTree::after);
	      }
	    else
	      { // Start at the beginning.
		first = realTree->first();
	      }
	    if (toObj)
	      { // Find the "next" item for this.  That is to say we stop
		// on the first item after this item, if it exists, or after
		// where this item would be, if it does not exist.
		pastEnd = realTree->find(toObj, objTree::before);
		if (pastEnd) { pastEnd = realTree->next(pastEnd); }
	      }
	    else
	      { // Stop at the end.
		pastEnd = NULL;
	      }
	  }

	for (objTreeIterator i = first; i != pastEnd; i = realTree->next(i))
	  {
	    if (!i)
	      { // This could only be the result of a bad compare function.
		// We have hit the end of the list before hitting the expected
		// last element.
		break;
	      }

	    // Save the key and the data, just for clarity.  Assume we already
	    // have a locked copy of each, based on having a locked copy
	    // of the parent structure.
	    Tcl_Obj *key = i->getData().getKey();
	    Tcl_Obj *data = i->getData().getData();

	    // Need to find the number of data items associated with this key.
	    int duplicates;
	    if (Tcl_ListObjLength(interp,
				  i->getData().getData(), 
				  &duplicates) != TCL_OK)
	      {
		return TCL_ERROR;
	      }
	    
	    for (int j = 0;j < duplicates; j++)
	      {
		if (keyVarObj)
		  {
		    if (!Tcl_ObjSetVar2(interp, keyVarObj, NULL, key,
					TCL_LEAVE_ERR_MSG | TCL_PARSE_PART1))
		      {
			return TCL_ERROR;
		      }
		  }
		if (dataVarObj)
		  { // Do a Tcl_ListObjIndex each time, rather than a
		    // Tcl_ListObjGetElements once, because the internal form
		    // of the list can change, even though it is locked, each
		    // time we execute arbitrary code.  We should not be able
		    // to get an error here, since we had a valid list above,
		    // but once we've gone this far the assertion doesn't cost
		    // us much.
		    Tcl_Obj *nextData;
		    if (Tcl_ListObjIndex(interp, data, j, &nextData) != TCL_OK)
		      {
			return TCL_ERROR;
		      }
		    // Make sure the current data value is properly garbage
		    // collected in case the SetVar doesn't work or does other
		    // strange things.
		    tclObject gc(nextData);
		    if (!Tcl_ObjSetVar2(interp, dataVarObj, NULL, 
					nextData,
					TCL_LEAVE_ERR_MSG | TCL_PARSE_PART1))
		      {
			return TCL_ERROR;
		      }
		  }
		int result = Tcl_EvalObj(interp, bodyObj);
		if (result == TCL_BREAK) { goto break_from_effective_loop; }
		if ((result != TCL_OK) && (result != TCL_CONTINUE))
		  {
		    return result;
		  }
	      }
	  }
	// The two C++ for loops make one appearent foreach loop.  Break
	// jumps out of the foreach, and therefore out of both fors.
      break_from_effective_loop:

	Tcl_ResetResult(interp);
	return TCL_OK;
      }
    case MMAP_KEYS:
      {
	if (objc != 3)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], false);
	if (!tree) { return TCL_ERROR; }
	objTree *realTree = &tree->getTree();
	tclObject result = Tcl_NewObj();
	for (objTreeIterator i = realTree->first(); i; i = realTree->next(i))
	  {
	    // Need to find the key...
	    Tcl_Obj *key = i->getData().getKey();

	    // ...and the number of identical keys.
	    int duplicates;
	    if (Tcl_ListObjLength(interp,
				  i->getData().getData(), 
				  &duplicates) != TCL_OK)
	      {
		return TCL_ERROR;
	      }
	    
	    for (;duplicates > 0; duplicates--)
	      {
		Tcl_ListObjAppendElement(NULL, result, key);
	      }
	  }
	Tcl_SetObjResult(interp, result);
	return TCL_OK;
      }
    case MMAP_DATA:
      {
	if (objc != 3)
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], true);
	if (!tree) { return TCL_ERROR; }
	Tcl_Obj *result = Tcl_NewObj();
	for (objTreeIterator i = tree->getTree().first(); 
	     i; 
	     i = tree->getTree().next(i))
	  {
	    Tcl_ListObjAppendList(NULL, result, i->getData().getData());
	  }
	Tcl_SetObjResult(interp, result);
	return TCL_OK;
      }
    case MMAP_VALUE:
      {
	if ((objc < 4) || (objc > 5))
	  {
	    Tcl_WrongNumArgs(interp, 2, objv, "var key ?error_value?");
	    return TCL_ERROR;
	  }
	tree = getTree(interp, objv[2], false);
	if (!tree) { return TCL_ERROR; }
	Tcl_Obj *found = tree->valueCmd(objv[3]);
	if (found)
	  { // Found!
	    int count;
	    if (Tcl_ListObjLength(interp, found, &count)
		!= TCL_OK) 
	      {
		return TCL_ERROR;
	      }
	    if (count)
	      { // This should always be the case, unless someone has modified
		// the tree by other means.  With this if we let a key mapping
		// to the empty list look like a non-existant key.
		Tcl_Obj *result;
		Tcl_ListObjIndex(NULL,  
				 found, 
				 count - 1,
				 &result);
		Tcl_SetObjResult(interp, result);
		return TCL_OK;
	      }
	  }
	// Not found in tree.
	if (objc == 5)
	  { // Use default.
	    Tcl_SetObjResult(interp, objv[4]);
	    return TCL_OK;
	  }
	else
	  { // Signal an error.
	    Tcl_AppendResult(interp, "No such element in tree.", NULL);
	    return TCL_ERROR;
	  }
      }    
    }
  return TCL_OK;
}
  

//--------------------------------------------------------------------------
// 
//--------------------------------------------------------------------------

extern "C" int Rbtree_Init(Tcl_Interp *interp) 
{
  tclTrue = Tcl_NewIntObj(1);
  Tcl_IncrRefCount(tclTrue);
  tclFalse = Tcl_NewIntObj(0);
  Tcl_IncrRefCount(tclFalse);

  Tcl_CreateObjCommand(interp, "tree::create", tree_create_cmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "tree::previous", 
		       tree_previous_cmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "tree::next", tree_next_cmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "tree::closest", tree_closest_cmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "tree::map", tree_map_cmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "tree::set", tree_set_cmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "tree::multiset",
		       tree_multiset_cmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "tree::multimap", 
		       tree_multimap_cmd, NULL, NULL);

  Tcl_RegisterObjType(&treeType);
  return TCL_OK;
}
